/******************************************************************************
NeuralNetBP filter plugin for Avisynth by V.C.Mohan
NeuralNetBP filter trains on input clip startframe and trainer clip tframe, then 
processes input frames. Uses back propogation. Classification type.
Author V.C.Mohan
Oct 30, 2005
********************************************************************************/  
//#include "windows.h"
//#include "avisynth.h"
//#include <cmath>

		



class NeuralNetBP : public GenericVideoFilter {
		PClip trainer;
		int nl;		//start frame number of  (child) clip. This and tframe will be used for training
		int enl;	// end frame number for this line
		int tframe;	// frame number of trainer clip to use for training
		 
	
		int slx;	// start x coordinate of rectangle left 
		int sty;	// start y coordinate of rectangle top  
		int srx;	// start x coord of window right 
		int sby;	// start y coord of window bottom 
		int elx;	// end x coordinate of rectangle left 
		int ety;	// end y coordinate of rectangle top  
		int erx;	// end x coord of window right 
		int eby;	// end y coord	of window bottom 
		bool white;	// Object to be marked as white or black.?
		int xpts;	// number of points along x axis to be used.. Only odd numbers 1 to 121
		int ypts;	// number of points along y axis to be used. Only odd numbers 1 to 121
		int tlx, tty, trx, tby;	// trainer clip window coordinates
		int thresh;	// thrshold. %age of maximum to be considered as object.
		double hf;		// hidden layer weight factor
		double of;		// output layer weight factor
		int iter;	// number of iterations for training
		bool test;	// is this test mode?
		double mstep;	// minimum step used by dividing with hf (learning rate)
		int wset;	// wset  weights are this set (random numbers start after skipping wset-1 * total weights) 
		int cwt;	// extra weight for correct value (acceleration)
	
		int hnodes;		// hidden layer nodes
		int best;		// use best weights out of this number of different starting weights
		const char * activate;	// name of activation function
		bool line;			// training window line or rectangle
	
		unsigned char *input;	// buffer to hold input values each time plus one bias
		double  *wh, *wo,*hlayer, *whcopy, *wocopy;	// weights for each node of hidden layer, output layer and output of hidden layer
		bool exist;			// flag for creation by new
		enum option {LOGISTIC =0, ALOGISTIC, TANH, ATANH,GAUSS, ELLIOT, POLY, EVERO};
		option opt;
		double *errbuf, *hinput;
		bool trainingOver;
		int *hist;			// histogram buffer
		int inodes;	//	(xpts*ypts+1)
		double oinput;	// the input (weighted sum) to the output node 
  public:
							//Definition of function
    NeuralNetBP(PClip _child,PClip _trainer,
				int _nl,int _enl,
				
				int _slx,int _sty,int _srx, int _sby,
				int _elx,int _ety,int _erx, int _eby,
				
				bool _white, int _xpts, int _ypts,
				int _tlx, int _tty, int _trx, int _tby,
				int _thresh,int _tframe,
				double _hf, double _of, int _iter,
				bool _test,double _mstep, 
				int _wset, int _cwt,
				
				int _hnodes, int _best,
				const char * _activate,
				bool _line,
				IScriptEnvironment* env) ;	
			
	~NeuralNetBP();				//destructor

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	void  getcase(const unsigned char * fp, int fpitch,int kb, int h,int w);
	double	getoutput();
//	double  getWeightedSum(double* inptr, double* inwts,int innodes);
//	double  getWeightedSum(unsigned char * inptr, double* inwts,int innodes);
//	double  passThroughSigmoid( double sum);
	void adjustWeights(double out, double errs, int nodes);

// This is the function that AviSynth calls to get a given frame.
// So when this functions gets called, the filter is supposed to return frame n.
};	
/*************************************************
 * The following is the implementation 
 * of the defined functions.
 *************************************************/
//Here is the acutal constructor code used

NeuralNetBP::NeuralNetBP(PClip _child,PClip _trainer,
			    int _nl,int _enl,
				
				int _slx,int _sty,int _srx, int _sby,
				int _elx,int _ety,int _erx, int _eby,
				
				bool _white, int _xpts, int _ypts,
				int _tlx, int _tty, int _trx, int _tby,
				int _thresh,int _tframe,
				double _hf, double _of, int _iter,
				bool _test,double _mstep,
				int _wset, int _cwt,
				
				int _hnodes, int _best,
				const char * _activate,
				bool _line,
				IScriptEnvironment* env) :

	GenericVideoFilter(_child),trainer(_trainer),
						nl(_nl),enl(_enl),
			
						slx(_slx), sty(_sty),
						srx(_srx), sby(_sby),
						elx(_elx), ety(_ety),
						erx(_erx), eby(_eby),
						white(_white),
						xpts(_xpts), ypts(_ypts),
						tlx(_tlx), tty(_tty),
						trx(_trx), tby(_tby),
						thresh(_thresh),
						tframe(_tframe),
						hf(_hf), of(_of), iter(_iter),
						test(_test),mstep(_mstep),
						wset(_wset), cwt(_cwt),
						
						hnodes(_hnodes),best(_best),
						activate(_activate),
						line (_line)

						
					
						
						 
			
{
	if(strcmp(activate, "logistic") ==0)
		opt = LOGISTIC;
	else if(strcmp(activate, "alogistic") ==0)
		opt = ALOGISTIC;
	else if(strcmp(activate, "tanh") ==0)
		opt = TANH;
	else if(strcmp(activate, "atanh") ==0)
		opt = ATANH;
	else if(strcmp(activate, "gauss") ==0)
		opt = GAUSS;
	else if(strcmp(activate, "elliot") ==0)
		opt = ELLIOT;
	else if(strcmp(activate, "evero") ==0)
		opt = EVERO;
	oinput=0.0;
	exist = false;
	trainingOver = false;
	inodes= xpts*ypts+1;

}

 NeuralNetBP::~NeuralNetBP()
{
	 if(exist)
	 {
		 delete []wh;	// hidden layer weights
		 delete []wo;	// output layer weights
		 delete []hlayer;	// output of hidden layer (input of output layer)
		 delete []whcopy;
		 delete []wocopy;
		 delete []input;
		 delete []errbuf;
		 delete []hinput;
		 delete []hist;
	 }

	
}

/***************************************************************/	
	PVideoFrame __stdcall NeuralNetBP::GetFrame(int in, IScriptEnvironment* env)
	{

		PVideoFrame Frame;
		int nframes= enl-nl;	// actual intervals
		int n;
		if(test)
		{
			Frame = child->GetFrame(nl, env);	//  process this frame only and show enl times
			
			if(in>=enl)
			{
				
				return Frame;
			}

			n=in;
			thresh =(100*n)/enl;
		}
		else
		{
	
			Frame = child->GetFrame(in, env);
			if(in<nl ||in>enl )
			
				return Frame;
			
			
			n=in-nl;
		}
		
		

		const unsigned char *fp= Frame->GetReadPtr();
		
		const	int bwd = vi.width;
        const	int bht = Frame->GetHeight();		
		const	int fpitch = Frame->GetPitch();
		int kb=vi.BitsPerPixel()/8;

		PVideoFrame work = env->NewVideoFrame(vi);
		unsigned char* wp =work->GetWritePtr();
		const	int wpitch =work->GetPitch();

		
		int object= white?255:0;
		int backg= white?0:255;

				// calculate parameters for this frame
		
		int lx=test?slx:slx+((elx-slx)*n)/nframes;
		int ty= test? sty:sty+((ety-sty)*n)/nframes;
		int by= test? sby:sby+((eby-sby)*n)/nframes;
		int rx= test? srx:srx+((erx-srx)*n)/nframes;
		
		
		

		if(vi.IsRGB())
		{
			// so that same code can be used for all color formats
			by=bht-1-(sty+((ety-sty)*n)/nframes);
			ty=bht-1-(sby+((eby-sby)*n)/nframes);
			int temp=bht-1-tby;
			tby = bht-1-tty;
			tty= temp;

		}
		
		int iterscale= iter>bwd-10  ? iter/(bwd-10) + 1 : 1;
		

		int vscale = bht>200? 1 : 200 /bht;	// scale for plotting 0 to 100 values (error and histogram)
		double emax=0.0;		// max error corresponding to 100% stored here
		
						;
		
		if(n==0 && ! trainingOver)
		{
			
			unsigned char * wpp= wp;
			// zero out output frame for drawing error and normalized histogram
			for(int h=0; h<bht; h++)
			{
				for(int w=0; w<bwd; w++)				
					wpp[kb*w] = backg;
				wpp+=wpitch;
			}
				// training is required
			PVideoFrame trframe = trainer->GetFrame(tframe, env);
			const unsigned char *tp= trframe->GetReadPtr();
			const	int tpitch = trframe->GetPitch();
			int xcases=(trx-tlx-xpts-1);
			int ycases=(tby-tty-ypts-1);
			int cases= xcases*ycases;
			
			if( !exist)
			{
				wh= new double[hnodes*inodes];		// buffer to hold hidden layer weights plus  bias
				wo= new double[hnodes];			// buffer to hold output layer weights
				hlayer = new double[hnodes];			// buffer for hidden layer output
				whcopy= new double[hnodes*inodes];		// buffer to hold best hidden layer weights plus  bias
				wocopy= new double[hnodes];			// buffer to hold best output layer weights
				errbuf = new double[iter/iterscale+1];
				hinput = new double[hnodes];		// To hold summed inputs to the hidden layer nodes
				input = new unsigned char[inodes];	// buffer to hold input plus one bias node
				hist = new int[256];
				exist=true;
			}

				// skip (wset-1)* total number weights random numbers
			for(int i=0; i<(hnodes*inodes +hnodes)*(wset-1); i++)
				rand();
				// setup for line input
			int number;
			double xinc, yinc;

			if(line)
			{
				if(abs(tby- tty) > abs(trx-tlx))					
					number = abs(tby-tty )+1;	// number of patterns  for training
				else
					number = abs(trx-tlx )+1;

				xinc = (trx-tlx)/(double)number;
				yinc = (tby-tty)/(double) number;
					
			}
				// repeat training with different start weights "best" of times	
			for(int b=0; b<best;b++)
			{
				// fill buffers with random weights

				for(int i=0; i<hnodes;i++)				
					// output layer weights values between -0.5 to +0.5
					wo[i]= (((double)rand())/RAND_MAX -0.5)*of;


				for( i=0;i<hnodes*inodes; i++)
					// hidden layer weights between -0.0005 and 0.0005
					wh[i]= (((double)rand())/RAND_MAX -0.5)*hf;


				
					// for tanh, atanh and elliot outputs -1 to 1, while logistic, alogistic, EVERO and gauss 0 to 1
				double zero = (opt== LOGISTIC || opt == ALOGISTIC || opt == GAUSS || opt == EVERO)? 0.0 : -1.0;

				// now calculate output and error

				for(i=0;i<iter;i++)		
				{
					double esum=0.0;
					double minesum;
					
					if(line)
					{
						for(int num=0; num<number;num++)
						{
							int h= tty+num*yinc;
							int w = tlx+num * xinc;
								// get next case for training
							getcase(fp, fpitch, kb, h,w);		// get next case input data + bias in input buffer
								// get output of the network with current weights after passing thru sigmoid filter
						
							double output=getoutput();
								// for tanh, atanh and elliot outputs -1 to 1, while logistic, alogistic and gauss 0 to 1
							double correct= *(tp+(h+ypts/2)*tpitch+kb*(w+xpts/2))==object?1.0:zero;
							double error=(opt== LOGISTIC || opt == ALOGISTIC || opt == GAUSS || opt == EVERO)?
								correct-output:(correct-output)/2;
							esum=esum+error*error;
							if(*(tp+(h+ypts/2)*tpitch+kb*(w+xpts/2))==object)
								error*=cwt;		// for enhancing weight for an accelerated solution, 
												//if you want more weight for object recognition 
												// rather than recognising non objects. 
							
								// adjust weights so that output value comes  closer to desired value
							adjustWeights(output,error, inodes);
							
						}
					}
					else
					{

						for(int h=tty; h< tby-ypts+1; h++)
							for(int w= tlx; w<trx-xpts+1; w++)
							{
							

							// get next case for training
								getcase(fp, fpitch, kb, h,w);		// get next case input data + bias in input buffer
							
								double output=getoutput();	
								// for tanh, atanh and elliot outputs -1 to 1, while logistic, alogistic and gauss 0 to 1
								double correct= *(tp+(h+ypts/2)*tpitch+kb*(w+xpts/2))==object?1.0:zero;
								double error=(opt== LOGISTIC || opt == ALOGISTIC || opt == GAUSS || opt == EVERO)?
								correct-output:(correct-output)/2;
								esum=esum+error*error;
								if(*(tp+(h+ypts/2)*tpitch+kb*(w+xpts/2))==object)
									error*=cwt;		// for enhancing weight for an accelerated solution, 
												//if you want more weight for object recognition 
												// rather than recognising non objects. 
							
								// adjust weights so that output value comes  closer to desired value
								adjustWeights(output,error, inodes);
							
							} 
					}
					
					
					emax= emax < esum? esum:emax;

							// to keep track of lowest esum acheived
					if(i==0 && b==0)
					
						minesum=esum;
							// if so far this is the best copy the weights for future use.
					if(minesum > esum)
					{
						for(int nodes=0; nodes<hnodes*inodes; nodes++)
							whcopy[nodes]=wh[nodes];
						for( nodes=0; nodes<hnodes; nodes++)
							wocopy[nodes] = wo[nodes];
						minesum=esum;
					}

						// error sum using the final random weights are stored. //
						//Note this may not be least. Otherwise require so many buffers
					if(b== best-1 && (i % iterscale)==0)
						errbuf[i/iterscale]= esum;


				}	// iterations
			
			}	// best of

				// copy weights got for lowest esum

			for(int nodes=0; nodes<hnodes*inodes; nodes++)
				wh[nodes]=whcopy[nodes];
			for( nodes=0; nodes<hnodes; nodes++)
				wo[nodes] = wocopy[nodes];

			trainingOver = true;
		
		}	// training end

		
		// as first frame output  histogram of processed test frame and training error plots

		if(n==0)
		{
			PVideoFrame work1 = env->NewVideoFrame(vi);
			unsigned char* wp1 =work1->GetWritePtr();
			const	int wpitch1 =work1->GetPitch();

			int vscale = bht>200? 1 : 200 /bht;
			// first frame histogram
			
			// zero out buffer
			for ( int nn=0;nn<256; nn++)
				hist[nn]=backg;
			
			for(int h=ty; h< by-ypts+1; h++)
				for(int w= lx; w<rx-xpts+1; w++)
				{

					// create histogram
					getcase(fp, fpitch, kb, h,w);		// get input data + bias in input buffer
				/* get output of the network with current weights after passing thru sigmoid filter
	
							for ( int hn=0; hn<hnodes; hn++)
							{
								hinput[hn]= getWeightedSum(input , wh+hn*inodes, inodes);
								hlayer[hn]= passThroughSigmoid(hinput[hn]);
							}

							double netIn= getWeightedSum(hlayer,wo, hnodes);
							double output=passThroughSigmoid(netIn);

				*/	double output=getoutput();	
					unsigned char pop = (opt== LOGISTIC || opt == ALOGISTIC 
										|| opt == GAUSS || opt == EVERO)?
										output*256 :(1+output)*128;
					hist[pop]++;
				}


			int hmax=0;
				// get max value for scaling
			for (  nn=0;nn<256; nn++)
				hmax=hist[nn]>hmax? hist[nn] : hmax;

			// now plot the histogram

			for(nn=0; nn<256;nn++)
				for(int val=0; val < (hist[nn]*100)/(hmax * vscale); val++)
					wp[(bht-1 -val)*wpitch + kb*(5+nn)] = object;	// indented 5 in width for visibility

			// plot horizontal scale for convenience
			for(nn=0; nn<256;nn+=25)
				for(int sht=0;sht<10;sht++)
					wp[(bht-1-sht)*wpitch + kb*(5+nn)]=128;

				
			// plot the error per iteration
			for(int i=0; i< iter/iterscale; i++)
			{
				int er= 100*errbuf[i] / emax; 
				wp[(bht/2-er/vscale)*wpitch + kb*(10+i)]=object;

			}
				// plot vertical scale

			for(nn=0; nn<11;nn++)
				for(int sht=0;sht<10;sht++)
				{
					wp[(bht/2-(nn*10)/vscale)*wpitch + kb*sht]=128;
					wp[(bht/2-(nn*10)/vscale)*wpitch + kb*(bwd-1-sht)]=128;
				}
				// plot iter scale

			for(nn=0; nn<iter/iterscale; nn+=(iter/iterscale)/10)
				for(int sht=0;sht<5;sht++)
				{
				
					wp[(bht/2-sht)*wpitch+kb*(10+nn)]=128;
					
				}

					// plot training window/ line

			
				

			if(line)
			{
				int number;
				double xinc, yinc;
				if(abs(tby- tty) > abs(trx-tlx))					
					number = abs(tby-tty )+1;	// number of patterns  for training
				else
					number = abs(trx-tlx )+1;

				xinc = (trx-tlx)/(double)number;
				yinc = (tby-tty)/(double) number;

				for(int num =0; num<number; num++)
					*(wp+ (tty+(int)(num*yinc))*wpitch + kb*(tlx+(int)(num*xinc)))=128;
					
			}

			else
			{
				for(int h=tty; h<tby; h++)
				{
					*(wp+h*wpitch + kb* tlx)=128;
					*(wp+h*wpitch + kb* trx)=128;
				}

				for(int w=tlx; w<trx; w++)
				{
					*(wp+tty*wpitch + kb* w)=128;
					*(wp+tby*wpitch + kb* w)=128;
				}
			}


							
		}
		else
		{
			
			// process regular

			if(opt== LOGISTIC || opt == ALOGISTIC || opt == GAUSS || opt == EVERO)
			{
			
				for(int h=ty; h< by-ypts+1; h++)
					for(int w= lx; w<rx-xpts+1; w++)
					{
						getcase(fp, fpitch, kb, h,w);		// get input data + bias in input buffer
					/* get output of the network with best weights and after passing thru sigmoid filter

							for ( int hn=0; hn<hnodes; hn++)
							{
								hinput[hn]= getWeightedSum(input , wh+hn*inodes, inodes);
								hlayer[hn]= passThroughSigmoid(hinput[hn]);
							}

							double netIn= getWeightedSum(hlayer,wo, hnodes);
							double output=passThroughSigmoid(netIn);
					*/	double output=getoutput();	
							
						if(output*100 >  thresh)
							*(wp+h*wpitch+kb*w)=object;
						else
							*(wp+h*wpitch+kb*w)=backg;

							
					}
			}

			else
			{
				// for tanh, atanh, elliot which output -1 to 1
				for(int h=ty; h< by-ypts+1; h++)
					for(int w= lx; w<rx-xpts+1; w++)
					{
						getcase(fp, fpitch, kb, h,w);		// get input data + bias in input buffer


					/* get output of the network with current weights after passing thru sigmoid filter
							for ( int hn=0; hn<hnodes; hn++)
							{
								hinput[hn]= getWeightedSum(input , wh+hn*inodes, inodes);
								hlayer[hn]= passThroughSigmoid(hinput[hn]);
							}

							double netIn= getWeightedSum(hlayer,wo, hnodes);
							double output=(1+passThroughSigmoid(netIn))/2;
								// get output of the network with best weights and after passing thru sigmoid filter
					*/	double output=(1+getoutput())/2;	
							
						if(output*100 >  thresh)
							*(wp+(h+ypts/2)*wpitch+kb*(w+xpts/2))=object;
						else
							*(wp+(h+ypts/2)*wpitch+kb*(w+xpts/2))=backg;

							
					}
			}

		}

		
				// grey border

			for(int h=ty;h<by;h++)
			{
				*(wp+h*wpitch+kb*lx)=128;
				*(wp+h*wpitch+kb*rx)=128;
			}

			for(int w=lx; w<rx; w++)
			{
				*(wp+ty*wpitch+kb*w)=128;
				*(wp+by*wpitch+kb*w)=128;
			}

			
				//  grey  u and v channels
			if(vi.IsYUY2())
			{
				for(h=ty;h<by;h++)
					for(w=lx;w<rx;w++)						
						*(wp+h*wpitch+kb*w+1)=128;
			}

			else if(vi.IsYV12())
			{
				
				unsigned char * wpU=work->GetWritePtr(PLANAR_U);
				unsigned char * wpV=work->GetWritePtr(PLANAR_V);
				const int  wpitchUV=work->GetPitch(PLANAR_U);

				for(h=ty/2;h<by/2;h++)
					for(w=lx/2;w<rx/2;w++)
					{
						// grey u and v channels
						*(wpU+h*wpitchUV+w)=128;
						*(wpV+h*wpitchUV+w)=128;
					}
			}

			else if(vi.IsRGB())
			{
				for(h=ty;h<by;h++)
					for(w=lx;w<rx;w++)
					{
								// grey by making  red green  blue equal .
						*(wp+h*wpitch+kb*w+1)=*(wp+h*wpitch+kb*w+2)=*(wp+h*wpitch+kb*w);
					}
			}



	return work;	
	
}

	

double	NeuralNetBP::getoutput()
					{
						switch (opt)
						{
							case LOGISTIC :
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
									hinput[hn]=sum;	

									// pass through sigmoid filter
									hlayer[hn]=1.0/(1.0+exp(-sum));
								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];

								// pass summed value through sigmoid to get output
								return ( 1.0/(1.0+exp(-sum)));
							}

							case ALOGISTIC :
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
									hinput[hn]=sum;
										
									sum/= 4.1;
									// pass through approximated sigmoid filter
									if(sum >=1)

										hlayer[hn]=1.0;

									else if (sum<=-1.0)
										hlayer[hn]=0;

									else if(sum<0)
										hlayer[hn]=0.5 + sum*(1.0+ sum/2);
									else
										hlayer[hn]=0.5 + sum*(1.0- sum/2);

								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];
								sum/= 4.1;

									// pass summed value through sigmoid to get output
								if(sum >=1)

										return(1.0);

									else if (sum<=-1.0)
										return(0);

									else if (sum<0)
										return(0.5 + sum*(1.0+sum/2));
									else
										return(0.5 + sum*(1.0- sum/2));

							}

							case TANH :
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
									hinput[hn]=sum;	
									
									// pass through tanh sigmoid filter
									
									hlayer[hn]=2.0/(1+exp(-2*sum))-1;

								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];

									// pass summed value through tanh sigmoid to get output
								
								return(2.0/(1+exp(-2*sum))-1);

							}

							case ATANH :
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
									hinput[hn]=sum;	
									
									// pass through approximate tanh sigmoid filter

									if(sum>1.92033)
										hlayer[hn]=0.96016;
									else if( sum <= -1.92033)
										hlayer[hn]= -0.96016;
									else if(sum >0)
										hlayer[hn]=0.96016 - 0.26037*(sum-1.92033)*(sum-1.92033);
									else
										hlayer[hn]= 0.26037*(sum+1.92033)*(sum+1.92033)-0.96016 ;

								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];

									// pass summed value through approximate tanh sigmoid to get output
								
								if(sum>1.92033)
										return (0.96016);
									else if( sum <= -1.92033)
										return ( -0.96016);
									else if(sum >0)
										return (0.96016 - 0.26037*(sum-1.92033)*(sum-1.92033));
									else
										return ( 0.26037*(sum+1.92033)*(sum+1.92033)-0.96016);

							}

							case GAUSS :
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
										
									hinput[hn]=sum;
									// pass through approximate gauss sigmoid filter

									
										hlayer[hn]=exp(-sum*sum);

								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];

									// pass summed value through gauss sigmoid to get output
								
								oinput=sum;
								return (exp(-sum*sum) );

							}

							case ELLIOT :
								
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
										
									
									// pass through  elliot sigmoid filter 1/(1+|X|)

									
									hlayer[hn]= sum >=0? sum/(1+sum) :sum/(1-sum);

								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];

									// pass summed value through elliot sigmoid to get output
								
								oinput=sum;
								return (sum >=0? sum/(1+sum) :sum/(1-sum));

							}

							case EVERO :
								
							{
								for ( int hn=0; hn<hnodes; hn++)		// loop for each hidden node
								{
									double sum=0.0;
									for ( int in=0; in<inodes; in++)
									{
										sum+= input[in]*wh[hn*(inodes)+in];
									}
										
									
									// pass through  EVERO sigmoid filter (X/2)/(1+|X|)+0.5

									
									hlayer[hn]= sum >=0? (sum/2)/(1+sum)+0.5 :(sum/2)/(1-sum)+0.5;

								}
						

									// find output value

								double sum=0.0;		// clear value

								for(hn=0; hn<hnodes; hn++)
									sum+=hlayer[hn]*wo[hn];

									// pass summed value through elliot sigmoid to get output
								
								oinput=sum;
								return (sum >=0? (sum/2)/(1+sum)+0.5 :(sum/2)/(1-sum)+0.5);

							}


							default: return (0);

						
						}
					}


void  NeuralNetBP::getcase(const unsigned char * fp, int fpitch,int kb, int h,int w)
{
	for(int i=0; i< ypts; i++)
		for(int n=0; n<xpts; n++)
			input[i*xpts+n]= *(fp+(h+i)*fpitch+kb*(w+n));
		input[xpts*ypts]=1;	// bias node value
}


void  NeuralNetBP::adjustWeights(double output, double error, int inodes)
{
	switch (opt)
	{
		case LOGISTIC :
		case ALOGISTIC:
		
		{
			double slopeo= output*(1.0-output);	// logistic slope function is y(1-y)
			for (int hn=0; hn<hnodes; hn++)		// loop for each hidden node
			{
			
				double slopeh= hlayer[hn] * (1.0-hlayer[hn]);

				for ( int in=0; in<inodes; in++)
				{
										
					double dxdw= input[in]*slopeh*wo[hn]*slopeo;
					wh[hn*inodes+in]= wh[hn*inodes+in]+ error*dxdw*mstep;
								
				}

			// find new weights for output layer
				double dxdw= hlayer[hn]*slopeo;
				wo[hn]=wo[hn]+dxdw*error*mstep;
			}

			break;
		}

		
		

		case TANH :
		case ATANH:
		{
			double slopeo= 1.0 - output*output;	// tanh slope function is 1-y*y
			for (int hn=0; hn<hnodes; hn++)		// loop for each hidden node
			{
				
				double slopeh=  1.0 - hlayer[hn] *hlayer[hn];

				for ( int in=0; in<inodes; in++)
				{
										
					double dxdw= input[in]*slopeh*wo[hn]*slopeo;
					wh[hn*inodes+in]= wh[hn*inodes+in]+ error*dxdw*mstep;
								
				}
							// find new weights for output layer
			
				double dxdw= hlayer[hn]*slopeo;
				wo[hn]=wo[hn]+dxdw*error*mstep;
			}
			break;
		}

		case GAUSS :
		{
			// Gauss slope function is -2*x*y.  input x, output y
			double slopeo= - 2.0 * output* oinput; // output sigmoid slope

			for (int hn=0; hn<hnodes; hn++)		// loop for each hidden node
			{
				
				
				double slopeh= -2*hinput[hn]*hlayer[hn];	// across hidden layer slope

				for ( int in=0; in<inodes; in++)
				{
										
					double dxdw= input[in]*slopeh*wo[hn]*slopeo;
					wh[hn*inodes+in]= wh[hn*inodes+in]+ error*dxdw*mstep;
								
				}
							// find new weights for output layer
			
			
				double dxdw= hlayer[hn]*slopeo;
				wo[hn]=wo[hn]+dxdw*error*mstep;
			}
			break;
		}

		case ELLIOT :
		{
			double slopeo;
				if(oinput>=0)
					slopeo=1/ ((1+ oinput)*(1+ oinput));	// Elliot slope function is 1/((1+|x|)*(1+|x|))
				else
					slopeo= 1/ ((1- oinput)*(1- oinput));
			
			for (int hn=0; hn<hnodes; hn++)		// loop for each hidden node
			{

			
				double slopeh=hinput[hn]>=0?
								  1/((1+hinput[hn])*(1+hinput[hn])):
									1/((1-hinput[hn])*(1-hinput[hn]));
			
				for ( int in=0; in<inodes; in++)
				{
					 
				
					double dxdw= input[in]*slopeh*wo[hn]*slopeo;
					wh[hn*inodes+in]= wh[hn*inodes+in]+ error*dxdw*mstep;
								
				}
							// find new weights for output layer
			
			
				double dxdw= hlayer[hn]*slopeo;
				wo[hn]=wo[hn]+dxdw*error*mstep;
			}
			break;
		}

		case EVERO :
		{
			double slopeo;
				if(oinput>=0)
					slopeo=1/ (2*(1+ oinput)*(1+ oinput))+0.5;	// EVERO slope function is 1/(2*(1+|x|)^2)
				else
					slopeo= 1/ (2*(1- oinput)*(1- oinput))+0.5;

			
			for (int hn=0; hn<hnodes; hn++)		// loop for each hidden node
			{
				

				double slopeh= hinput[hn]>=0? 1/(2*(1+hinput[hn])*(1+hinput[hn])) + 0.5:
												1/(2*(1-hinput[hn])*(1-hinput[hn])) + 0.5;
			
				for ( int in=0; in<inodes; in++)
				{
					
										
					double dxdw= input[in]*slopeh*wo[hn]*slopeo;
					wh[hn*inodes+in]= wh[hn*inodes+in]+ error*dxdw*mstep;
								
				}
							// find new weights for output layer
			
			
				double dxdw= hlayer[hn]*slopeo;
				wo[hn]=wo[hn]+dxdw*error*mstep;
			}
			break;
		}
	}
}


/***************************************************************/
// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,

AVSValue __cdecl Create_NeuralNetBP(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	if(!args[0].AsClip()->GetVideoInfo().IsRGB24() && !args[0].AsClip()->GetVideoInfo().IsRGB32() 
		&& !args[0].AsClip()->GetVideoInfo().IsYUY2() && !args[0].AsClip()->GetVideoInfo().IsYV12())
		env->ThrowError("NeuralNetBP: This color format is not supported here");
	int nf=args[0].AsClip()->GetVideoInfo().num_frames;
	if(args[2].AsInt(0)<0 || args[2].AsInt(0)> args[3].AsInt(nf-1))
		env->ThrowError("NeuralNetBP: Frame number must be positive and end frame not less than start frame");
	if(args[2].AsInt(0)>=nf || args[3].AsInt(nf-1)>=nf)
		env->ThrowError("NeuralNetBP:this clip which has only %d frames",nf);
	if(!args[0].AsClip()->GetVideoInfo().pixel_type != !args[1].AsClip()->GetVideoInfo().pixel_type
		|| !args[0].AsClip()->GetVideoInfo().width != !args[1].AsClip()->GetVideoInfo().width
		|| !args[0].AsClip()->GetVideoInfo().height != !args[1].AsClip()->GetVideoInfo().height)
		env->ThrowError("NeuralNetBP:The input clip and trainer clip have differing color formats or dimensions");
	if(args[1].AsClip()->GetVideoInfo().num_frames<=args[20].AsInt(0))
		env->ThrowError("NeuralNetBP:the training clip does not have the specified frame");
	
	int xx=args[0].AsClip()->GetVideoInfo().width-1;
	int yy=args[0].AsClip()->GetVideoInfo().height-1;
	
	
	if(args[4].AsInt(0)<0 || args[5].AsInt(0) <0 || args[8].AsInt(0)<0 || args[9].AsInt(0)<0)
		env->ThrowError("NeuralNetBP: coordinates should not be -ve");
	if(args[6].AsInt(xx)<args[4].AsInt(0)+args[13].AsInt(9) || args[7].AsInt(yy) <args[5].AsInt(0)+args[14].AsInt(9)
		|| args[6].AsInt(xx)>xx || args[7].AsInt(yy)>yy)		
		env->ThrowError("NeuralNetBP: window x and y dimensions should not be less than  xpts or ypts and must be within frame");

	
	if(args[10].AsInt(xx)<args[8].AsInt(args[4].AsInt(0))+args[13].AsInt(9) 
		|| args[11].AsInt(args[7].AsInt(yy)) <args[9].AsInt(args[5].AsInt(0))+args[14].AsInt(9)
		|| args[10].AsInt(args[6].AsInt(xx))>xx || args[11].AsInt(args[7].AsInt(yy))>yy)		
		env->ThrowError("NeuralNetBP: end window x and y dimensions should not be less than  xpts or ypts and must be within frame");
	

	if(args[13].AsInt(9) <1 || args[14].AsInt(9)<1 || args[13].AsInt(9)*args[14].AsInt(9)>121 
		|| args[13].AsInt(9)*args[14].AsInt(9)<9|| (args[13].AsInt(9) & 1) ==0 || (args[14].AsInt(9) & 1) ==0)
		env->ThrowError("NeuralNetBP:xpts and ypts must be odd numbers 1 to 121, and xpts * ypts must be 9 to 121");

	int wx, wy;
	if(args[31].AsBool(true))
	{
		wx=xx-args[13].AsInt(9);
		wy= yy -args[14].AsInt(9);
		if(args[15].AsInt(args[4].AsInt(0)) <0 || args[15].AsInt(args[4].AsInt(0)) > wx
			|| args[17].AsInt(wx) <0 || args[17].AsInt(wx) > wx)
			env->ThrowError("NeuralNetRP: The training line x coordinates must be within frame and less than frame width- xpts");

		if(args[16].AsInt(args[5].AsInt(0)) <0 || args[16].AsInt(args[5].AsInt(0)) > wy
			|| args[18].AsInt(wy) <0 || args[18].AsInt(wy) > wy)
			env->ThrowError("NeuralNetRP: The training line y coordinates must be within frame and less than frame height- ypts");
	}

	else
	{
		wx=xx;
		wy=yy;

		if(args[15].AsInt(args[4].AsInt(0)) <0 || args[15].AsInt(args[4].AsInt(0))>args[17].AsInt(wx)
		|| args[17].AsInt(wx) > wx)
			env->ThrowError("NeuralNetRP: training window x coordinates be +ve, within frame ");

		if(args[16].AsInt(args[5].AsInt(0)) <0 || args[16].AsInt(args[5].AsInt(0))>args[18].AsInt(wy)
		|| args[18].AsInt(wy) > wy)
			env->ThrowError("NeuralNetRP: The  training window y coordinates be +ve, within frame ");
	}
	
	if(args[19].AsInt(60) <5 || args[19].AsInt(60) >95)
		env->ThrowError("NeuralNetBP:threshold %age value must be above 5 and 95");
	if(args[21].AsFloat(0.0005) <=0 || args[22].AsFloat(1.0) <=0)
		env->ThrowError("NeuralNetBP:the weight factors must be positive numbers");
	if(args[23].AsInt(3200) <=0)
		env->ThrowError("NeuralNetBP:number of iterations must be positive number");
	if(args[25].AsFloat(0.00005) <=0)
		env->ThrowError("NeuralNetBP:minstep must not be 0 or negative");
	if(args[27].AsInt(5) <=0 || args[27].AsInt(5) > 25)
		env->ThrowError("NeuralNetBP:extra weight cwt must be 1 to 25 only");
	if(args[28].AsInt(10) <sqrt(args[13].AsInt(9) * args[14].AsInt(9)))
		env->ThrowError("NeuralNetBP:Too few hidden nodes? suggest minimum of %d", (int)sqrt(args[13].AsInt(9) * args[14].AsInt(9)));
	if(args[29].AsInt(1)<0 || args[29].AsInt(1)>10)
		env->ThrowError("NeuralNetBP:bestof can be between 1 and 10 only");
	if(args[26].AsInt(1)<0 || args[26].AsInt(1)>10)
		env->ThrowError("NeuralNetBP:wset can be between 1 and 10 only");
	
	if( strcmp(args[30].AsString("logistic"), "logistic") !=0 && strcmp(args[30].AsString("logistic"), "alogistic") !=0
		&& strcmp(args[30].AsString("logistic"), "tanh") !=0 && strcmp(args[30].AsString("logistic"), "atanh") !=0
		&& strcmp(args[30].AsString("logistic"), "gauss") !=0 && strcmp(args[30].AsString("logistic"), "elliot") !=0
		&& strcmp(args[30].AsString("logistic"), "evero") !=0)
		env->ThrowError("NeuralNetBP:Un implemented activation function specified");

		return  new NeuralNetBP(	args[0].AsClip(),			// clip for input
								args[1].AsClip(),			// training clip
								args[2].AsInt(0),			// start frame number to process						
								args[3].AsInt(nf-1),		// end frame number to process
								
								args[4].AsInt(0),			// rectangle start left x, 
								args[5].AsInt(0),			// rectangle start top y,  
								args[6].AsInt(xx),			// start right x 
								args[7].AsInt(yy),			// start bottom y  
								args[8].AsInt(args[4].AsInt(0)),		// rectangle end left x, 
								args[9].AsInt(args[5].AsInt(0)),		// rectangle end top y,  
								args[10].AsInt(args[6].AsInt(xx)),		// end rightx 
								args[11].AsInt(args[7].AsInt(yy)),		// end bot y
								
								args[12].AsBool(true),		// white as object?

								args[13].AsInt(9),			// input points in x direction
								args[14].AsInt(9),			// input points in y direction
								args[15].AsInt(args[4].AsInt(0)),	// training clip window left x
								args[16].AsInt(args[5].AsInt(0)),	// training clip window top y
								args[17].AsInt(wx), // training clip window right x
								args[18].AsInt(wy), // training clip window bottom y
								args[19].AsInt(60),		// threshold : %age value of max to be taken as object
								args[20].AsInt(0),			// trainer clip frame number to use
								args[21].AsFloat(0.0005),	// hidden layer weights factor
								args[22].AsFloat(1.0),		// output layer weights factor
								args[23].AsInt(3200),	// number of iterations
								args[24].AsBool(false),	// is this a test?
								args[25].AsFloat(0.00005),		// minimum step . this divided by hidden layer weight factor is used
								args[26].AsInt(1),			// weight set number (start wts are this set)
								args[27].AsInt(5),				// extra weightage for correct value
								
								
								args[28].AsInt(10),				// hidden layer nodes
								args[29].AsInt(1),				// use best (least error) weights out of this many different starting weights.
								args[30].AsString("logistic"),	// type of activation function to use
								args[31].AsBool(true),			// training along diagonal line or rectangle?
							
								env);
	



	
}
// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.
/*
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("NeuralNetBP", "cc[sf]i[ef]i[lx]i[ty]i[rx]i[by]i[elx]i[ety]i[erx]i[eby]i"
		"[white]b[xpts]i[ypts]i[tlx]i[tty]i[trx]i[tby]i[thresh]i[tf]i[wh]f[wo]f[iter]i"
		"[test]b[minstep]f[wset]i[cwt]i[hnodes]i[bestof]i[activate]s[line]b", Create_NeuralNetBP, 0);
					
    return "NeuralNetBP  Plugin";
}
*/
